/*
KiFi, KDE Wireless Manager
Copyright (c) 2002-2004 by Chris TenHarmsel <chris@staticmethod.net>

KiFi, (c) 2004 by the KiFi developers

*************************************************************************
*                                                                       *
* This program is free software; you can redistribute it and/or modify  *
* it under the terms of the GNU General Public License as published by  *
* the Free Software Foundation; either version 2 of the License, or     *
* (at your option) any later version.                                   *
*                                                                       *
*************************************************************************
*/

#include <kpassivepopup.h>
#include <klocale.h>
#include <kdebug.h>
#include <qtimer.h>
#include <qcstring.h>
#include <qpushbutton.h>
#include <klistbox.h>
#include <kinputdialog.h>
#include <qmessagebox.h>
#include <qframe.h>

#include <string.h>

#include "kifisystraycontroller.h"
#include "ui/kifinetworkchooserdialog.h"
#include "wirelessinfoengine.h"
#include "kifisystemtray.h"
#include "networkinterface.h"
#include "wirelessnetworkscanner.h"

KiFiSystrayController::KiFiSystrayController(
    QString interface, QObject *parent, const char* name )
		: QObject( parent, name ),
		_interfaceName( interface ),
		_currentlyScanning( false ),
_knownNetworks() {
	// Create the wireless info engine, don't need to delete
	_engine = new WirelessInfoEngine( _interfaceName, this );

	// Create the system tray
	_sysTray = new KiFiSystemTray();
	// Create a network scanner, don't need to delete
	_scanner = new WirelessNetworkScanner( _interfaceName, this );

	// initialize the network chooser dialog
	_chooserDialog = new KiFiNetworkChooserDialog( 0L );

	// Create the timer for updating the wireless network info
	_infoTimer = new QTimer();

	// Initialize the signal connections
	initConnections();

	// Tell the timer to start running, fire every 5 seconds
	_infoTimer->start( 5000 );

	// Set the update interval
	_scanner->setUpdateInterval( 3000 );

	kdDebug() << "Showing the system tray for network: " << _interfaceName << endl;
	// Show the systray icon
	_sysTray->show();
}

KiFiSystrayController::~KiFiSystrayController() {
	// Stop the timer
	_infoTimer->stop();

	// Stop the scanner
	_scanner->stopScanning();

	// delete the network chooser dialog
	delete _chooserDialog;

	// Clean up after ourselves
	delete _sysTray;
	delete _infoTimer;
}

void KiFiSystrayController::initConnections() {
	// New network
	connect( _scanner, SIGNAL( networkFound( WirelessNetworkInfo & ) ),
	         this, SLOT( slotNetworkFound( WirelessNetworkInfo & ) ) );
	// Network update
	connect( _scanner, SIGNAL( networkUpdated( WirelessNetworkInfo & ) ),
	         this, SLOT( slotNetworkUpdated( WirelessNetworkInfo & ) ) );
	// Network lost
	connect( _scanner, SIGNAL( networkLost( WirelessNetworkInfo & ) ),
	         this, SLOT( slotNetworkLost( WirelessNetworkInfo & ) ) );

	// Connect the timer to the handler for it
	connect( _infoTimer, SIGNAL( timeout() ),
	         this, SLOT( slotInfoTimerHandler() ) );

	// Connect the user controls for stopping and starting scanning
	connect( _sysTray, SIGNAL( startScanningSelected() ),
	         this, SLOT( slotStartScanning() ) );
	connect( _sysTray, SIGNAL( stopScanningSelected() ),
	         this, SLOT( slotStopScanning() ) );

	// Connect the user controls for the quit menu option
	connect( _sysTray, SIGNAL( quitSelected() ),
	         this, SLOT( slotHandleQuit() ) );

	// Connect the user controls for the single click event
	connect( _sysTray, SIGNAL( infoSelected() ),
	         this, SLOT( slotShowInformationWindow() ) );

	// Connect the Use button clicked handler for the
	// network chooser dialog.
	connect( _chooserDialog, SIGNAL( useNetwork( const QString & ) ),
	         this, SLOT( slotUseSelectedNetwork( const QString & ) ) );
	connect( _chooserDialog, SIGNAL( cancelSelected() ),
	         this, SLOT( slotCloseNetworkChooser() ) );
	connect( _chooserDialog, SIGNAL( networkHighlighted( const QString& ) ),
	         this, SLOT( slotUpdateNetworkInfoDetails( const QString& ) ) );
}

/*!
    \fn KiFiSystrayController::slotUpdateNetworkInfoDetails(QString &)
 */
void KiFiSystrayController::slotUpdateNetworkInfoDetails( const QString & text ) {
	_chooserDialog->showDetailedInfo( _knownNetworks[ text ] );
}

void KiFiSystrayController::slotStartScanning() {
	if ( !_currentlyScanning ) {
		kdDebug() << "Checking to see if the interface is up" << endl;
		NetworkInterface iface( _interfaceName );
		bool isUp = iface.isUp();
		if ( !isUp ) {
			kdDebug() << "Interface is down, bringing up to start scanning" << endl;
			// Bring the interface at least up
			bool success = iface.up();
			if ( !success ) {
				kdDebug() << "Could not bring interface up to start scanning" << endl;
				return ;
			}
		}

		kdDebug() << "Starting scanning for wireless networks on "
		<< _interfaceName << endl;
		// Start the scanner
		_scanner->startScanning();
		_currentlyScanning = true;
	}
}

void KiFiSystrayController::slotStopScanning() {
	if ( _currentlyScanning ) {
		kdDebug() << "Stopping scanning for wireless networks on "
		<< _interfaceName << endl;
		// Start the scanner
		_scanner->stopScanning();
		_currentlyScanning = false;
	}
}

void KiFiSystrayController::slotInfoTimerHandler() {
	int quality = _engine->getSignalQuality();
	QString ssid;

	kdDebug() << "Got signal quality of: " << quality << endl;

	// Get the SSID we're connected to
	if ( _engine->updateInfo() ) {
		ssid = _engine->getSSID();
	}

	_sysTray->setSignalStrengthDisplay( quality, ssid );

}

void KiFiSystrayController::slotShowInformationWindow() {
	if ( ! _chooserDialog->isShown() ) {
		_chooserDialog->show();
	}
}

void KiFiSystrayController::slotNetworkFound( WirelessNetworkInfo & info ) {
	// Show a passive popup, not quite sure why we multiply the signal strength by two
	// but it gives more consistant results with the rest of the system
	KPassivePopup::message(
	    i18n( "Access Point Found: " ) + info.getSSID(),
	    i18n( "Signal Strength: " ) + QString::number( info.getSignalStrength() * 2 ),
	    _sysTray );
	// Add this new interface to our know networks
	_knownNetworks.insert( info.getSSID(), info );
	kdDebug() << "Adding network to chooser dialog" << endl;
	_chooserDialog->addNetwork( info.getSSID() );
}

void KiFiSystrayController::slotNetworkUpdated( WirelessNetworkInfo & info ) {
	// Update the information in the map, or rather insert it with
	// the overwrite flag on (default)
	_knownNetworks.insert( info.getSSID(), info );
	// Find out which network is currently being viewed in the chooser dialog
	// and update the info if it's the one that just got updated
	QString chosenNetwork = _chooserDialog->getSelectedItem();
	if ( chosenNetwork = info.getSSID() ) {
		_chooserDialog->showDetailedInfo( info );
	}
}

void KiFiSystrayController::slotNetworkLost( WirelessNetworkInfo & info ) {
	// Show a passive popup
	KPassivePopup::message( i18n( "Access Point Lost: " ) + info.getSSID(),
	                        _sysTray );
	// Remove this network from the list of available networks for this
	// interface
	_knownNetworks.remove( info.getSSID() );
	kdDebug() << "Removing network from chooser dialog" << endl;
	_chooserDialog->removeNetwork( info.getSSID() );
}

void KiFiSystrayController::slotHandleQuit() {
	kdDebug() << "KiFiSystrayController for " << _interfaceName
	<< ": user selected quit" << endl;
	emit quitSelected();
}

void KiFiSystrayController::slotCloseNetworkChooser() {
	kdDebug() << "KiFiSytrayController::slotCloseNetworkChooser()" << endl;
	_chooserDialog->hide();
}

void KiFiSystrayController::slotUseSelectedNetwork( const QString & networkName ) {

	WirelessNetworkInfo info = _knownNetworks[ networkName ];

	kdDebug() << "KiFiSystrayController::useSelectedNetwork()" << endl;

	kdDebug() << "Switching to network: " << info.getSSID() << endl;
	kdDebug() << "             Address: " << info.getAddress() << endl;
	kdDebug() << "                 Wep: " << info.getWepRequired() << endl;
	kdDebug() << "     Signal Strength: " << info.getSignalStrength() << endl;

	QString wepKey = QString::null;

	NetworkInterface iface( _interfaceName );
	iface.configureDHCP();
	iface.stop();

	kdDebug() << "Setting Mode on the interface" << endl;
	if ( !iface.setMode( info.getMode() ) ) {
		kdDebug() << "Error setting mode on wireless card" << endl;
	} else {
		kdDebug() << "Finished setting mode" << endl;
	}

	// Set the SSID
	kdDebug() << "Setting SSID on the interface" << endl;
	if ( !iface.setEssid( info.getSSID() ) ) {
		kdDebug() << "Error setting SSID on wireless card" << endl;
	} else {
		kdDebug() << "Finished setting SSID" << endl;
	}

	// set WEP
	if ( info.getWepRequired() ) {
		kdDebug() << "WEP Required, prompting for key and setting" << endl;
		wepKey = KInputDialog::getText( i18n( "WEP Key" ), i18n( "Enter WEP Key" ) );
		if ( !iface.setWepRequired( wepKey ) ) {
			kdDebug() << "Error setting SSID on wireless card" << endl;
		} else {
			kdDebug() << "Finished setting WEP key on card" << endl;
		}
	} else {
		kdDebug() << "WEP NOT required, setting card to open" << endl;
		if ( !iface.setWepNotRequired() ) {
			kdDebug() << "Error setting no wep key on wireless card" << endl;
		} else {
			kdDebug() << "Finished setting no wep key on wireless card" << endl;
		}
	}

	// Commit the info on the card
	kdDebug() << "Attempting to commit wireless info" << endl;
	if ( !iface.commitWirelessInfo() ) {
		kdDebug() << "Could not commit wireless info to the card" << endl;
	} else {
		kdDebug() << "Wirless info committed" << endl;
	}

	// Restart the interface
	if ( iface.start() ) {
		kdDebug() << _interfaceName << " brought up successfully" << endl;
		// Close the network chooser
		slotCloseNetworkChooser();
		QMessageBox::information( _sysTray, i18n( "Interface Restarted" ),
		                          i18n( "Joined new network: " ) + info.getSSID(), i18n( "Ok" ) );
	} else {
		kdDebug() << _interfaceName << " could not be brought up" << endl;
		QMessageBox::information( _sysTray, i18n( "Error" ),
		                          i18n( "Could not join network: " ) + info.getSSID(), i18n( "Ok" ) );
	}

}

#include "kifisystraycontroller.moc"

/*
 * Local variables:
 * c-indentation-style: k&r
 * c-basic-offset: 2
 * indent-tabs-mode: t
 * End:
 */