/*
	KiFi, KDE Wireless Manager
	Copyright (c) 2002-2004 by Chris TenHarmsel <chris@staticmethod.net>

	KiFi, (c) 2004 by the KiFi developers

	*************************************************************************
	*                                                                       *
	* This program is free software; you can redistribute it and/or modify  *
	* it under the terms of the GNU General Public License as published by  *
	* the Free Software Foundation; either version 2 of the License, or     *
	* (at your option) any later version.                                   *
	*                                                                       *
	*************************************************************************
	*/

#include "wirelessnetworkscanner.h"
#include <kdebug.h>
#include <qtimer.h>
#include <qstringlist.h>

#include <linux/rtnetlink.h>
#include <net/ethernet.h>
#include <linux/netlink.h>

WirelessNetworkScanner::WirelessNetworkScanner(
	QString & iface, QObject *parent, const char *name)
	: QObject(parent, name ),
		_interface( iface ),
		_knownNetworks()
{
	_timer = new QTimer( this );
	// Connect the timer to the update method
	QObject::connect( _timer, SIGNAL( timeout() ),
										this, SLOT( updateInfo() ) );
	_timerInterval = 5000;

	_iwsocket = iw_sockets_open();
}

WirelessNetworkScanner::~WirelessNetworkScanner()
{
	// TODO
}

void WirelessNetworkScanner::setUpdateInterval( int msec )
{
	// Dictate a 200ms floor
	if ( msec > 200 ) {
		_timerInterval = msec;
	}
}

void WirelessNetworkScanner::startScanning()
{
	// If the timer is not currently running
	if ( !_timer->isActive() ) {
		// Start it
		_timer->start(_timerInterval, false );
	}
}

void WirelessNetworkScanner::stopScanning()
{
	if ( _timer->isActive() ) {
		// Stop it
		_timer->stop();
	}
}

void WirelessNetworkScanner::updateInfo()
{

	// NOTE: Much of this code was directly taken from iwlist.c

	// Flag to indicate whether or not we got good data
	bool haveGoodData = false;

	// iwreq is found in wireless.h, this is taken from
	// setup in iwlist.c
	struct iwreq wrq;
	wrq.u.param.flags = IW_SCAN_DEFAULT;
  wrq.u.param.value = 0;

	// Results
  char buffer[IW_SCAN_MAX_DATA];
  struct iw_range range;
  bool has_range;

	// Select timeout
  struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 250000;

	// 5s timeout
  int timeout = 1000000;

	// Get range stuff
  has_range =
		( iw_get_range_info(_iwsocket, _interface.latin1(), &range) >= 0 );

	if ( !has_range ) {
		kdDebug() << k_funcinfo << "Interface does not have range support" << endl;
		return;
	} else if ( range.we_version_compiled < 14 ) {
		kdDebug() << k_funcinfo << "Wireless Extensions version < 14" << endl;
		return;
	}

	// Possible that we have to use an rtnetlink socket (see manpages)
	// instead of the normal socket that we usually use for talking to
	// the wireless card

	//	kdDebug() << "Creating rtnetlink socket" << endl;
	//	int rtnetlink_socket = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);

	kdDebug() << "Asking driver to begin scanning" << endl;
	// Tell the driver to start scanning
  if( iw_set_ext(
				_iwsocket, _interface.latin1(), SIOCSIWSCAN, &wrq) < 0) {
		// This is an error case
		if(errno != EPERM) {
			kdDebug() << k_funcinfo << "Interface " << _interface
								<< " doesn't support scanning : " << strerror(errno) << endl;
			return;
		}

		kdDebug() << "SIOCSIWSCAN IOCTL call returned error: " << strerror( errno )
							<< endl;
		// If we don't have the permission to initiate the scan, we may
		// still have permission to read left-over results.
		// But, don't wait !!!
		tv.tv_usec = 0;
	}

	// Update the overall timeout, not sure why this is done
	timeout -= tv.tv_usec;

	kdDebug() << "Beginning loop waiting for information from the driver "
						<< "to be ready" << endl;
	// Begin looping, waiting for information from the card to be ready
	while(1) {
		// File descriptors for select
		fd_set rfds;
		// Last fd
		int ret;

		// Guess what ? We must re-generate rfds each time
		FD_ZERO(&rfds);
		// Put the socket into the list
		FD_SET( _iwsocket, &rfds );

		kdDebug() << "Calling select to wait for socket activity, timing out after: "
			<< tv.tv_usec << "usecs" << endl;
		// Wait until something happens on the socket
		ret = select(1, &rfds, NULL, NULL, &tv);

		/* Check if there was an error */
		if(ret < 0) {
			// If it's an acceptable error, and we have time left
			if(( errno == EAGAIN || errno == EINTR ) && ( timeout > 0 )) {
				kdDebug() << "Got error from select: " << strerror( errno )
									<< ".  Non-fatal, continuing loop" << endl;
				// go back to waiting with the select statment
				continue;
			} else {
				kdDebug() << "Unhandled signal while waiting on select" << endl;
				break;
			}
		} else {
			kdDebug() << "Select returned successfully, attempting to retrieve data"
								<< endl;
			// Try to read the results
			wrq.u.data.pointer = buffer;
			wrq.u.data.flags = 0;
			wrq.u.data.length = sizeof(buffer);

			// SIOCGIWSCAN does a GET basically
			if(iw_get_ext(
					 _iwsocket, _interface.latin1(), SIOCGIWSCAN, &wrq) < 0) {
				// Error case
				// Check if results not available yet
				if(errno == EAGAIN) {
					// Restart timer for only 100ms, for some reason they get reset
					tv.tv_sec = 0;
					tv.tv_usec = 100000;
					// Update the overall timeout
					timeout -= tv.tv_usec;
					// If overall timeout is expired, break out
					if(timeout < 0){
						break;
					}
					kdDebug() << "Failed to retrieve info, trying again" << endl;
					kdDebug() << "Time left: " << timeout << "usecs" << endl;
				} else {
					// This is a Bad error
					kdDebug() << _interface << ": Failed to read scan data : "
										<< strerror(errno) << endl;
					break;
				}
			} else {
				kdDebug() << "Found good scan data, breaking out of read loop" << endl;
				// We have the results, go to process them
				haveGoodData = true;
				break;
			}
		}

		// ORIGINAL COMMENT
		// In here, check if event and event type
		// if scan event, read results. All errors bad & no reset timeout
	}
	// END LOOP

	// By now, wrq should contain some data
	if ( haveGoodData && wrq.u.data.length ) {
		kdDebug() << "Have good data, attempting to parse" << endl;
		struct iw_event iwe;
		struct stream_descr stream;
		int ap_num = -1;
		int ret;
		char essid[IW_ESSID_MAX_SIZE + 1];
		typedef QMap<int, WirelessNetworkInfo> NetworkMap;
		NetworkMap networks;

		kdDebug() << "Initializing event stream..." << endl;
		// Initialize the event stream
		// Events are not per AP found.  It's like every piece of info about
		// each AP found classifies as an "event" and they could come in any order
		iw_init_event_stream( &stream, buffer, wrq.u.data.length );

		// Parse each event
		do{
			// Extract an event and print it
			ret = iw_extract_event_stream(&stream, &iwe, range.we_version_compiled);
			if (ret > 0) {
				switch (iwe.cmd)
				{
					case SIOCGIWAP: {
						// This means that we are starting to display info from a new AP
						ap_num++;
						kdDebug() << "Found new AP number " << ap_num << endl;
						// Create a new entry in the map
						networks[ap_num] = WirelessNetworkInfo();
						// address to sprintf address data into
						QString address;
						// ethernet address data
						struct ether_addr * addr =
							(struct ether_addr *) &iwe.u.ap_addr.sa_data;
						// set the address to a mac-address style string
						networks[ap_num].setAddress(
								address.sprintf("%02X:%02X:%02X:%02X:%02X:%02X",
								addr->ether_addr_octet[0], addr->ether_addr_octet[1],
								addr->ether_addr_octet[2], addr->ether_addr_octet[3],
								addr->ether_addr_octet[4], addr->ether_addr_octet[5]));
						break;
					}
					case SIOCGIWNWID: {
						// This is the Network ID, someone tell me what this is....
						if ( iwe.u.nwid.disabled ) {
							kdDebug() << "NWID: disabled" << endl;
							// Set the flag to false in the data object
							networks[ap_num].setNWIDEnabled( false );
							networks[ap_num].setNWID( -1 );
						} else {
							// Set the flag to true in the data object
							networks[ap_num].setNWIDEnabled( true );
							// Set the value via: iwe.u.nwid.value
							networks[ap_num].setNWID( iwe.u.nwid.value );
							kdDebug() << "NWID:" << iwe.u.nwid.value << endl;
						}
						break;
					}
					case SIOCGIWMODE: {
						QString mode = QString( iw_operation_mode[ iwe.u.mode ] );
						kdDebug() << "Mode: " << mode << endl;
						networks[ap_num].setMode( mode );
						break;
					}
					case SIOCGIWRATE: {
						// Buffer to hold the bitrate string, shouldn't need more then 25 EVER
						char bitrateBuf[25];
						iw_print_bitrate( bitrateBuf, 25, iwe.u.bitrate.value );
						QString bitrate = QString( bitrateBuf );
						networks[ap_num].addSupportedRate( bitrate );
						kdDebug() << "Rate: " << bitrate << endl;
						break;
					}
					case SIOCGIWESSID: {
						// SSID Event
						// Original code from iwlist:
						// ap_num = print_scanning_token(&iwe, ap_num, &range, has_range);
						if (iwe.u.essid.pointer && iwe.u.essid.length) {
							memcpy(essid, iwe.u.essid.pointer, iwe.u.essid.length * sizeof(char));
						}
						// Append a null so it's a valid string
						essid[iwe.u.essid.length] = '\0';
						kdDebug() << "SSID: " << essid << endl;
						networks[ap_num].setSSID( QString( essid ) );
						break;
					}
					case IWEVQUAL: {
						// Quality
						if ( iwe.u.qual.qual ) {
							kdDebug() << "Quality: " << ( int )iwe.u.qual.qual << endl;
							networks[ap_num].setSignalStrength( ( int )iwe.u.qual.qual );
						}
						break;
					}
					case SIOCGIWENCODE: {
						// Encoding info (ie. WEP)
						// Buffer to hold the key found
						unsigned char key[IW_ENCODING_TOKEN_MAX];

						// Figure out if there's a WEP key or not
						if(iwe.u.data.pointer) {
							memcpy(key, iwe.u.essid.pointer, iwe.u.data.length);
						} else{
							iwe.u.data.flags |= IW_ENCODE_NOKEY;
						}

						// If WEP is disabled
						if ( iwe.u.data.flags & IW_ENCODE_DISABLED ) {
							kdDebug() << "WEP Disabled" << endl;
							networks[ap_num].setWepRequired( false );
						} else {
							// otherwise there is an iwlib function to parse it for us
							// buffer is reusable I think

							// Original code from iwlist.c that shows the key?
							// iw_print_key( buffer, sizeof( buffer ), key, sizeof( key ),
							// iwe.u.data.flags );
							// Buffer should now contain the key

							kdDebug() << "WEP Enabled" << endl;
							networks[ap_num].setWepRequired( true );
							if((iwe.u.data.flags & IW_ENCODE_INDEX) > 1) {
								kdDebug() << ( int )( iwe.u.data.flags & IW_ENCODE_INDEX)
													<< endl;
							}
							if(iwe.u.data.flags & IW_ENCODE_RESTRICTED) {
								kdDebug() << "Security mode: restricted" << endl;
							}
							if(iwe.u.data.flags & IW_ENCODE_OPEN) {
								kdDebug() << "Security mode: open" << endl;
							}
						}
						break;
					}
					case SIOCGIWFREQ: {
						// Frequency/channel
						double freq;
						// Converted to channel
						int channel = -1;
						// iw_freq2float is from iwlib.h
						freq = iw_freq2float( &( iwe.u.freq ) );

						// Convert to channel if possible
						if ( has_range ) {
							channel = iw_freq_to_channel( freq, &range );
						}

						// Set stuff on the wireless network info object
						char freqString[32];
						iw_print_freq_value(freqString, 32, freq);
						networks[ap_num].setFrequency( freqString );
						networks[ap_num].setChannel( channel );
						kdDebug() << "Frequency: " << freqString << endl;
						kdDebug() << "Channel: " << channel << endl;
						break;
					}
					default: {
						kdDebug() << "Unknown Event: [" << iwe.cmd << "]"
											<< endl;
					}
				}
			} else {
				kdDebug() << "Error extracting wireles scanning event from stream: "
					<< strerror( errno ) << endl;
			}
		} while ( ret > 0 );

		// Print out all the wireless networks found
		kdDebug() << "################ SCANNING SUMMARY ################" << endl;
		// Build a map of SSID's to whether or not they're still
		// available, it'll be updated in the following loop
		QMap <QString, bool> networkFoundList;

		NetworkMap::Iterator it;
		for ( it = networks.begin(); it != networks.end(); ++it ) {
			kdDebug() << "Network " << it.key() << endl;
			kdDebug() << "\tSSID: " << it.data().getSSID() << endl;
			kdDebug() << "\tAddress: " << it.data().getAddress() << endl;
			kdDebug() << "\tQuality: " << it.data().getSignalStrength() << endl;
			kdDebug() << "\tWEP Required: " << it.data().getWepRequired() << endl;
			kdDebug() << "\tFrequency: " << it.data().getFrequency() << endl;
			kdDebug() << "\tChannel: " << it.data().getChannel() << endl;

			// Store it for the future
			kdDebug() << "Looking for [" << it.data().getSSID()
								<< "] in the known networks map" << endl;

			if ( !_knownNetworks.contains( it.data().getSSID() ) ) {
				kdDebug() << "Not found, adding" << endl;

				// Add this new networks to the known networks
				_knownNetworks[it.data().getSSID()] = it.data();

				// Emit new network found
				kdDebug() << "Emitting networkFound" << endl;

				emit networkFound( it.data() );

			} else {
				// We previously knew about this SSID, see if this one is the same
				WirelessNetworkInfo &known = _knownNetworks[it.data().getSSID()];
				WirelessNetworkInfo &newNetwork = it.data();

				if ( known == newNetwork ) {
					kdDebug() << "Network info did not change, not emitting signal"
										<< endl;
				} else {
					// Emit network updated
					kdDebug() << "Network info updated, emitting networkUpdated" << endl;
					emit networkUpdated( it.data() );

					// Update the known networks map with the new data
					_knownNetworks[it.data().getSSID()] = it.data();
				}
			}

			// Loop through all the ever known networks
			// Check to see if each is found in the available list
			// If one is not available anymore, but was previously known
			// then emit the networkLost signal, and remove it from the
			// known networks ivar
			QStringList networksToRemoveFromKnownList;

			kdDebug() << "Checking to see if any known networks were lost" << endl;
			QMap <QString, WirelessNetworkInfo>::Iterator it;
			for ( it = _knownNetworks.begin(); it != _knownNetworks.end(); it++ ) {
				bool found = false;
				// Check if the previously known network was just scanned.
				QMap <int,  WirelessNetworkInfo>::Iterator sit;
				for ( sit = networks.begin(); sit != networks.end(); sit++ ) {
					if ( it.data().getSSID() == sit.data().getSSID() ) {
						// we found it
						found = true;
						break;
					}
				}

				if ( !found ) {
					kdDebug() << "Network [" << it.data().getSSID() << "] was not found"
										<< ", marking it for removal" << endl;
					// Add it to the list of networks to remove
					networksToRemoveFromKnownList.push_back( it.data().getSSID() );
				}
			}

			// Remove networks that were not found on this scanned
			for ( size_t i = 0; i < networksToRemoveFromKnownList.size(); i++ ) {
				QString ssid = networksToRemoveFromKnownList[i];
				kdDebug() << "Emitting networkLost signal for ["
									<< ssid << "]" << endl;
								// Emit a signal saying this network was removed
				emit networkLost( _knownNetworks[ssid] );
				// Remove it from the known networks list
				_knownNetworks.remove( ssid );
			}

		}
	} else {
		kdDebug() << _interface << ": No scan results found" << endl;
		QMap <QString, WirelessNetworkInfo>::Iterator it;
		for ( it = _knownNetworks.begin(); it != _knownNetworks.end(); it++ ) {
			// Emit a signal that we're removing it
			emit networkLost( it.data() );
		}

		// Empty the known networks
		_knownNetworks.clear();
	}
	return;
}

void WirelessNetworkScanner::run()
{
	while ( 1 )
	{
		updateInfo();
		sleep( 5 );
	}
}

#include "wirelessnetworkscanner.moc"

/*
 * Local variables:
 * c-indentation-style: k&r
 * c-basic-offset: 2
 * indent-tabs-mode: t
 * End:
 */
